/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/* vim: set ts=2 sw=2 expandtab: */
/*
 * Copyright (c) 2009 Drexel University
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Tom Wambold <tom5760@gmail.com>
 */
/* These classes implement the NHDP neighbor discovery protocol.  See
 * http://tools.ietf.org/wg/manet/draft-ietf-manet-nhdp/ for more info */

#ifndef __NHDP_H__
#define __NHDP_H__

#include <map>
#include <queue>
#include <set>

#include "ns3/application.h"
#include "ns3/event-id.h"
#include "ns3/ptr.h"
#include "ns3/ipv4-address.h"
#include "ns3/socket.h"
#include "ns3/packetbb.h"
#include "ns3/random-variable.h"

#include "nhdp-info-base.h"

namespace ns3 {

/* PacketBB Types */
const uint8_t ADDR_TLV_LOCAL_IF = 1;

/* PacketBB Values */
const uint8_t ADDR_TLV_LOCAL_IF_THIS = 1;
const uint8_t ADDR_TLV_LOCAL_IF_OTHER = 2;

/** Used as the comparator when making heaps out of tuples with expiration
* times. */
template <class T>
class TimeCompare
{
public:
  bool operator() (const Ptr<T> &a, const Ptr<T> &b)
  {
    return b->time < a->time;
  }
};

/**
 * \ingroup applications 
 * \defgroup nhdp NHDP
 */

/**
 * \ingroup nhdp
 * \brief A NHDP client
 */
class NhdpClient : public Application 
{
public:
  typedef std::map<uint32_t, Ptr<NhdpLocalInterfaceTuple> >::const_iterator
    ConstLocalInterfaceIterator;

  NhdpClient (void);
  virtual ~NhdpClient (void);

  static TypeId GetTypeId (void);

  /* Local Interface Base Methods */
  /** Prevents interface from being used for transmitting HELLO messages */
  void MarkIfaceNonManet (uint32_t ifaddr);

  /* End Information Bases */

  /** 
   * \brief Adds a PacketBB message to be sent out with the next hello message.
   * \param message a PacketBB message to include
   */
  //void QueueMessage (Ptr<PbbMessage> message);

  /**
   * \brief Sets a callback for a particular message type.
   *
   * \param messageType the PbbMessage message type to listen for
   * \param cb the callback to call when a message is received
   *
   * On reception of a message of that type, the callback will be called.
   * Multiple callbacks can be registered to one message type.
   */
  //void RegisterMessageCallback (uint8_t messageType, Callback<PbbMessage> cb);

  void HandleRecv (Ptr<Socket> socket);

protected:
  virtual void DoDispose (void);

private:
  virtual void StartApplication (void);
  virtual void StopApplication (void);

  void ScheduleHello (Ptr<Socket> socket);
  void SendHello (Ptr<Socket> socket);

  //void CleanRemovedInterfaceAddressSet (void);

  Ptr<PbbAddressBlock> BuildLocalAddressBlock (Ptr<Socket> socket);

  /* Configuration paramters */
  Ipv4Address m_address;
  uint32_t m_port;

  Time m_helloInterval;
  Time m_helloMinInterval;
  Time m_refreshInterval;

  Time m_lHoldTime;
  Time m_hHoldTime;

  double m_hystAccept;
  double m_hystReject;
  double m_initialQuality;
  bool m_initialPending;

  Time m_hpMaxJitter;
  Time m_htMaxJitter;

  Time m_nHoldTime;
  Time m_iHoldTime;

  /* Other attributes */
  bool m_running;
  std::set<uint32_t> m_nonManetSet;
  RandomVariable m_rng;
  std::map<Ptr<Socket>, Ipv4Address> m_socketAddresses;


  /*
  std::map< uint32_t, Ptr<Socket> > m_indexSockets;
  std::map< Ptr<Socket>, uint32_t > m_socketsIndex;

  std::queue< Ptr<PbbMessage> > m_messages;
  std::multimap< uint8_t, Callback<PbbMessage> > m_callbacks;
  */
};

} /* namespace ns3 */

#endif /* __NHDP_H__ */
